==> python_elementstx-0.1.3-py3-none-any.whl: elementstx/liquid_addr.py <==
# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# pylama:ignore=E501,E261

"""Reference implementation for Blech32 and segwit addresses."""

from typing import List, Tuple, Optional, Union


CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def blech32_polymod(values: List[int]) -> int:
    """Internal function that computes the blech32 checksum."""
    generator = [0x7d52fba40bd886, 0x5e8dbf1a03950c, 0x1c3a3c74072a18, 0x385d72fa0e5139, 0x7093e5a608865b] # new generators, 7 bytes
    chk = 1
    for value in values:
        top = chk >> 55 # 25->55
        chk = ((chk & 0x7fffffffffffff) << 5) ^ value # 0x1ffffff->0x7fffffffffffff
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def blech32_hrp_expand(hrp: str) -> List[int]:
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


def blech32_verify_checksum(hrp: str, data: List[int]) -> bool:
    """Verify a checksum given HRP and converted data characters."""
    return blech32_polymod(blech32_hrp_expand(hrp) + data) == 1


def blech32_create_checksum(hrp: str, data: List[int]) -> List[int]:
    """Compute the checksum values given HRP and data."""
    values = blech32_hrp_expand(hrp) + data
    polymod = blech32_polymod(values + [0]*12) ^ 1 # 6->12
    return [(polymod >> 5 * (11 - i)) & 31 for i in range(12)]
#                            ^ 5                          ^ 6


def blech32_encode(hrp: str, data: List[int]) -> str:
    """Compute a blech32 string given HRP and data values."""
    combined = data + blech32_create_checksum(hrp, data)
    return hrp + '1' + ''.join([CHARSET[d] for d in combined])


def blech32_decode(bech: str) -> Tuple[Optional[str], Optional[List[int]]]:
    """Validate a blech32 string, and determine HRP and data."""
    if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
            (bech.lower() != bech and bech.upper() != bech)):
        return (None, None)
    bech = bech.lower()
    pos = bech.rfind('1')
    if pos < 1 or pos + 13 > len(bech) or len(bech) > 1000: # 7->13 90->1000
        return (None, None)
    if not all(x in CHARSET for x in bech[pos+1:]):
        return (None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos+1:]]
    if not blech32_verify_checksum(hrp, data):
        return (None, None)
    return (hrp, data[:-12]) # 6->12


def convertbits(data: Union[bytes, List[int]], frombits: int, tobits: int,
                pad: bool = True) -> Optional[List[int]]:
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def decode(hrp: str, addr: str) -> Tuple[Optional[int], Optional[bytes]]:
    """Decode a segwit confidential address.
    Its payload is longer than the payload for unconfidential address
    by 33 bytes (the length of blinding pubkey)"""
    hrpgot, data = blech32_decode(addr)
    if hrpgot != hrp:
        return (None, None)
    assert data is not None
    decoded = convertbits(data[1:], 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 40+33:
        return (None, None)
    if data[0] > 16:
        return (None, None)
    if data[0] == 0 and len(decoded) != 20+33 and len(decoded) != 32+33:
        return (None, None)
    return (data[0], bytes(decoded))


def encode(hrp: str, witver: int, witprog: bytes) -> Optional[str]:
    """Encode a segwit address."""
    decoded = convertbits(witprog, 8, 5)
    if decoded is None:
        return None
    ret = blech32_encode(hrp, [witver] + decoded)
    if decode(hrp, ret) == (None, None):
        return None
    return ret

==> python_elementstx-0.1.4-py3-none-any.whl: elementstx/liquid_addr.py <==
# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# pylama:ignore=E501

"""Reference implementation for Blech32 and segwit addresses."""

from typing import List, Tuple, Optional, Union

from enum import Enum


CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


BECH32_CONST = 1
BECH32M_CONST = 0x455972a3350f7a1


class Encoding(Enum):
    """Enumeration type to list the various supported encodings."""
    BECH32 = 1
    BECH32M = 2


def blech32_polymod(values: List[int]) -> int:
    """Internal function that computes the blech32 checksum."""
    generator = [0x7d52fba40bd886, 0x5e8dbf1a03950c, 0x1c3a3c74072a18, 0x385d72fa0e5139, 0x7093e5a608865b]  # new generators, 7 bytes
    chk = 1
    for value in values:
        top = chk >> 55  # 25->55
        chk = ((chk & 0x7fffffffffffff) << 5) ^ value  # 0x1ffffff->0x7fffffffffffff
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def blech32_hrp_expand(hrp: str) -> List[int]:
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


def blech32_verify_checksum(hrp: str, data: List[int]) -> Optional[Encoding]:
    """Verify a checksum given HRP and converted data characters."""
    check = blech32_polymod(blech32_hrp_expand(hrp) + data)
    if check == BECH32_CONST:
        return Encoding.BECH32
    elif check == BECH32M_CONST:
        return Encoding.BECH32M
    else:
        return None


def blech32_create_checksum(encoding: Encoding, hrp: str, data: List[int]
                            ) -> List[int]:
    """Compute the checksum values given HRP and data."""
    values = blech32_hrp_expand(hrp) + data
    const = BECH32M_CONST if encoding == Encoding.BECH32M else BECH32_CONST
    polymod = blech32_polymod(values + [0]*12) ^ const  # 6->12
    return [(polymod >> 5 * (11 - i)) & 31 for i in range(12)]
#                            ^ 5                          ^ 6


def blech32_encode(encoding: Encoding, hrp: str, data: List[int]) -> str:
    """Compute a blech32 string given HRP and data values."""
    combined = data + blech32_create_checksum(encoding, hrp, data)
    return hrp + '1' + ''.join([CHARSET[d] for d in combined])


def blech32_decode(bech: str) -> Tuple[Optional[Encoding], Optional[str], Optional[List[int]]]:
    """Validate a blech32 string, and determine HRP and data."""
    if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
            (bech.lower() != bech and bech.upper() != bech)):
        return (None, None, None)
    bech = bech.lower()
    pos = bech.rfind('1')
    if pos < 1 or pos + 13 > len(bech) or len(bech) > 1000:  # 7->13 90->1000
        return (None, None, None)
    if not all(x in CHARSET for x in bech[pos+1:]):
        return (None, None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos+1:]]
    encoding = blech32_verify_checksum(hrp, data)
    if encoding is None:
        return (None, None, None)
    return (encoding, hrp, data[:-12])  # 6->12


def convertbits(data: Union[bytes, List[int]], frombits: int, tobits: int,
                pad: bool = True) -> Optional[List[int]]:
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def decode(hrp: str, addr: str) -> Tuple[Optional[int], Optional[bytes]]:
    """Decode a segwit confidential address.
    Its payload is longer than the payload for unconfidential address
    by 33 bytes (the length of blinding pubkey)"""
    encoding, hrpgot, data = blech32_decode(addr)
    if hrpgot != hrp:
        return (None, None)
    assert data is not None
    decoded = convertbits(data[1:], 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 40+33:
        return (None, None)
    if data[0] > 16:
        return (None, None)
    if data[0] == 0 and len(decoded) != 20+33 and len(decoded) != 32+33:
        return (None, None)
    if (data[0] == 0 and encoding != Encoding.BECH32) or \
            (data[0] != 0 and encoding != Encoding.BECH32M):
        return (None, None)
    return (data[0], bytes(decoded))


def encode(hrp: str, witver: int, witprog: bytes) -> Optional[str]:
    """Encode a segwit address."""
    encoding = Encoding.BECH32 if witver == 0 else Encoding.BECH32M
    decoded = convertbits(witprog, 8, 5)
    if decoded is None:
        return None
    ret = blech32_encode(encoding, hrp, [witver] + decoded)
    if decode(hrp, ret) == (None, None):
        return None
    return ret
